include("Parameters.jl")

using DifferentialEquations
using Plots
using DataInterpolations

function Plant(Nw, U_opt, States, t_min)
    if Nw == 1
        par = WellPar1()
        @unpack_WellPar1 par
    elseif Nw == 2
        par = WellPar2()
        @unpack_WellPar2 par
    elseif Nw == 3
        par = WellPar3diff()
        @unpack_WellPar3diff par
    else
        println("Nw must be either 1 or 2 or 3!")
        #More data can be added in Parameters.jl 
    end

    ##function
    function Gas_Lift_Model(dx,x,p,t)
        ## mass flow rates
        ω_gl = p(t)
        m_ga = x[1:Nw, 1] # mass of gas in annulus
        m_gt = x[1:Nw, 2] # mass of gas in tubing
        m_ot = x[1:Nw, 3] # mass of oil in tubing
        ## Algabraic Equations
        P_a = 1e-5*(((R.*T_a./(V_a.*Mw) + 9.81.*H_a./V_a).*m_ga.*1e3) + (Mw./(R.*T_a).*((R.*T_a./(V_a.*Mw) + 9.81.*H_a./V_a).*m_ga.*1e3)).*9.81.*H_a) # annulus pressure
        ρ_a = 1e-2*(Mw./(R.*T_a).*P_a*1e5)  # gas, in annulus
        P_wh = 1e-5*(((R.*T_w./Mw).*(m_gt*1e3./(L_w.*A_w + L_bh.*A_bh .- m_ot.*1e3./ρₒᵢₗ))) - ((m_gt*1e3+m_ot*1e3 ) ./ (L_w.*A_w)).*9.81.*H_w./2) # wellhead pressure
        rho_m = 1e-2*(((m_gt*1e3 + m_ot*1e3).*P_wh*1e5.*Mw.*ρₒᵢₗ) ./ (m_ot*1e3.*P_wh*1e5.*Mw + ρₒᵢₗ.*R.*T_w.*m_gt*1e3))  # mixture, in tubing
        ω_pc = C_pc.*sqrt.(rho_m*1e2.*(P_wh*1e5 .- P_m*1e5)) # total flow through choke
        P_wi = 1e-5*((P_wh*1e5 + 9.81./(A_w .* L_w) .* (m_ot*1e3+m_gt*1e3 .- ρₒᵢₗ .* L_bh .* A_bh) .* H_w + 128*mu_oil .* L_w .* ω_pc ./(3.141*D_w .^ 4 .*((m_gt*1e3 + m_ot*1e3) .* P_wh * 1e5 .* Mw .* ρₒᵢₗ) ./ (m_ot*1e3 .* P_wh*1e5.*Mw + ρₒᵢₗ .* R .* T_w .* m_gt * 1e3)))) 
        ω_iv = C_iv .*sqrt.(ρ_a*1e2.*(P_a*1e5 - P_wi*1e5)) 
        ω_pg = (m_gt*1e3./(m_gt*1e3+m_ot*1e3)).*ω_pc # produced gas rate
        ω_po = (m_ot*1e3./(m_gt.*1e3+m_ot.*1e3)).*ω_pc # produced oil rate
        P_bh = 1e-5*(P_wi*1e5 .+ ρₒᵢₗ*9.81.*H_bh + 128*mu_oil.*L_bh.*ω_po./(3.14 * D_bh .^4 .* ρₒᵢₗ)) 
        ω_ro = (PI)*1e-6.*(Press_r*1e5 .- P_bh*1e5) 
        ω_rg = 1e1*GOR.*ω_ro 
        ## Differential Equations
        dx[1:Nw, 1] = (ω_gl - ω_iv)*1e-3 
        dx[1:Nw, 2] = (ω_iv + ω_rg*1e-1 - ω_pg)*1e-3 
        dx[1:Nw, 3] = (ω_ro - ω_po)*1e-3 
    end

    tspan = (0.0, t_min*60)
    t_points = collect(tspan[1]:4:tspan[2])

    gas_lift_1 = ConstantInterpolation(U_opt[1,:],t_points)
    gas_lift_2 = ConstantInterpolation(U_opt[2,:],t_points)
    gas_lift_3 = ConstantInterpolation(U_opt[3,:],t_points)

	p_fun(t) = [gas_lift_1(t), gas_lift_2(t), gas_lift_3(t)]

    # I assumed U_opt is [1:3,1:16]
    #Problem Loading
    prob = ODEProblem(Gas_Lift_Model, States[:x0], tspan, p_fun)

    #Solve
    sol = DifferentialEquations.solve(prob, BS3(),saveat=4)
    # sol = DifferentialEquations.solve(prob,saveat=4)
	#storing the trajectories
	temp_1 = Array(sol)[1,:,:]
    temp_2 = Array(sol)[2,:,:]
    temp_3 = Array(sol)[3,:,:]
    m_ga_profile = [temp_1[1,:]';temp_2[1,:]';temp_3[1,:]']
    m_gt_profile = [temp_1[2,:]';temp_2[2,:]';temp_3[2,:]']
    m_ot_profile = [temp_1[3,:]';temp_2[3,:]';temp_3[3,:]']


    U_profile = zeros(3,length(t_points))
    for i =1:length(t_points)
        U_profile[:,i] = p_fun(t_points[i])
    end
	
##==================================================
    # m_ga0 = sol.u[end][:, 1]
    # m_gt0 = sol.u[end][:, 2]
    # m_ot0 = sol.u[end][:, 3]
	state0 = [m_ga_profile[:,end] m_gt_profile[:,end] m_ot_profile[:,end]]
    # state0 = [temp_1[:,end] temp_2[:,end] temp_3[:,end]]
    ##
    # dxi =zeros(Nw,3)
    P_a = 1e-5*(((R.*T_a./(V_a.*Mw) + 9.81.*H_a./V_a).*m_ga_profile.*1e3) + (Mw./(R.*T_a).*((R.*T_a./(V_a.*Mw) + 9.81.*H_a./V_a).*m_ga_profile.*1e3)).*9.81.*H_a) # annulus pressure
    ρ_a = 1e-2*(Mw./(R.*T_a).*P_a*1e5)  # gas, in annulus
    P_wh = 1e-5*(((R.*T_w./Mw).*(m_gt_profile*1e3./(L_w.*A_w + L_bh.*A_bh .- m_ot_profile.*1e3./ρₒᵢₗ))) - ((m_gt_profile*1e3+m_ot_profile*1e3 ) ./ (L_w.*A_w)).*9.81.*H_w./2) # wellhead pressure
    rho_m = 1e-2*(((m_gt_profile*1e3 + m_ot_profile*1e3).*P_wh*1e5.*Mw.*ρₒᵢₗ) ./ (m_ot_profile*1e3.*P_wh*1e5.*Mw + ρₒᵢₗ.*R.*T_w.*m_gt_profile*1e3))  # mixture, in tubing
    ω_pc = C_pc.*sqrt.(rho_m*1e2.*(P_wh*1e5 .- P_m*1e5)) # total flow through choke
    P_wi = 1e-5*((P_wh*1e5 + 9.81./(A_w .* L_w) .* (m_ot_profile*1e3+m_gt_profile*1e3 .- ρₒᵢₗ .* L_bh .* A_bh) .* H_w + 128*mu_oil .* L_w .* ω_pc ./(3.141*D_w .^ 4 .*((m_gt_profile*1e3 + m_ot_profile*1e3) .* P_wh * 1e5 .* Mw .* ρₒᵢₗ) ./ (m_ot_profile*1e3 .* P_wh*1e5.*Mw + ρₒᵢₗ .* R .* T_w .* m_gt_profile * 1e3)))) 
    ω_iv = C_iv .*sqrt.(ρ_a*1e2.*(P_a*1e5 - P_wi*1e5)) 
    ω_pg = (m_gt_profile*1e3./(m_gt_profile*1e3+m_ot_profile*1e3)).*ω_pc # produced gas rate
    ω_po = (m_ot_profile*1e3./(m_gt_profile.*1e3+m_ot_profile.*1e3)).*ω_pc # produced oil rate
    P_bh = 1e-5*(P_wi*1e5 .+ ρₒᵢₗ*9.81.*H_bh + 128*mu_oil.*L_bh.*ω_po./(3.14 * D_bh .^4 .* ρₒᵢₗ)) 
    ω_ro = (PI)*1e-6.*(Press_r*1e5 .- P_bh*1e5) 
    ω_rg = 1e1*GOR.*ω_ro
    dx1 = (U_profile - ω_iv)*1e-3 
    dx2 = (ω_iv + ω_rg*1e-1 - ω_pg)*1e-3 
    dx3 = (ω_ro - ω_po)*1e-3

    myDict = Dict(
        :m_ga_pro   => m_ga_profile,
        :m_gt_pro   => m_gt_profile,
        :m_ot_pro   => m_ot_profile,
        :U_pro   => U_profile,
        :P_a0    => P_a,
        :ρ_a0    => ρ_a,
        :P_wh0   => P_wh,
        :ρ_m0    => rho_m,
        :ω_pc0   => ω_pc,
        :P_wi0   => P_wi,
        :ω_iv0   => ω_iv,
        :ω_pg0   => ω_pg,
        :ω_po0   => ω_po,
        :P_bh0   => P_bh,
        :ω_ro0   => ω_ro,
        :ω_rg0   => ω_rg,
        :dx0     => zeros(3,3),
        :x0      => state0,
        :dx1     => dx1,
        :dx2     => dx2,
        :dx3     => dx3,
    )
    
    return myDict #,m_ga_profile, m_gt_profile, m_ot_profile, U_profile
end